;; The first three lines of this file were inserted by DrRacket. They record metadata
;; about the language level of this file in a form that our tools can easily process.
#reader(lib "htdp-intermediate-lambda-reader.ss" "lang")((modname game) (read-case-sensitive #t) (teachpacks ((lib "image.rkt" "teachpack" "2htdp") (lib "universe.rkt" "teachpack" "2htdp"))) (htdp-settings #(#t constructor repeating-decimal #f #t none #f ((lib "image.rkt" "teachpack" "2htdp") (lib "universe.rkt" "teachpack" "2htdp")) #f)))
;; Paul Friederichsen
;; CSCI 1301 Game Project
;; v1.01

;; Repo and issues: https://github.com/floogulinc/csci1301-game

;; Travis CI testing: https://travis-ci.org/floogulinc/csci1301-game

;; Implemented features tracked here: https://github.com/floogulinc/csci1301-game/issues/4


(require "require.rkt") ; have to use a seperate file to require things because of using a student language

; path to audio files (windows can only play .wav)
(define jump-sound "res/smb_jump-super.wav")
(define coin-sound "res/smb_coin.wav")
(define pause-sound "res/smb_pause.wav") 

(define play-asynchronously #t) ; continue execution while file is playing?

(gui-play-sound coin-sound play-asynchronously) ; play startup coin sound

(define tileset-mario1 (bitmap/file "res/tileset-mario1.png")) ; mario tileset image


; crop-at-xy: number, number, image -> image
; crop s the given image to a 16x16 image at 16 times the x and y given
(define (crop-at-xy x y image) (crop (* 16 x) (* 16 y) 16 16 image))

(check-expect (crop-at-xy 0 0 tileset-mario1) (crop 0 0 16 16 tileset-mario1))
(check-expect (crop-at-xy 3 2 tileset-mario1) (crop 48 32 16 16 tileset-mario1))


; the list of tiles that can be used in a world
(define useable-tiles (list (crop-at-xy 0 0 tileset-mario1)
                            (crop-at-xy 1 0 tileset-mario1)
                            (crop-at-xy 2 0 tileset-mario1)
                            (crop-at-xy 3 0 tileset-mario1)
                            (crop-at-xy 24 0 tileset-mario1)
                            (crop-at-xy 0 1 tileset-mario1)))

(define num-useable-tiles (length useable-tiles)) ; gives the number of useable tiles


; a tile structure consists of:
; id: the id of the image for the tile
; x: the x location of the tile (in terms of 16x16 tiles)
; y: the y location of the tile (in terms of 16x16 tiles)
(define-struct tile (id x y))

; default tiles for the start state of the program
(define default-tiles (append (build-list 64 (lambda (x) (make-tile 0 x 14)))
                              (build-list 64 (lambda (x) (make-tile 0 x 15)))
                              (list (make-tile 3 11 13) (make-tile 3 12 13) (make-tile 3 13 13) (make-tile 3 14 13) (make-tile 3 15 13)
                                    (make-tile 3 12 12) (make-tile 3 13 12) (make-tile 3 14 12) (make-tile 3 15 12)
                                    (make-tile 3 13 11) (make-tile 3 14 11) (make-tile 3 15 11)
                                    (make-tile 3 14 10) (make-tile 3 15 10)
                                    (make-tile 3 15 9)
                                    (make-tile 4 20 9)
                                    (make-tile 1 24 9)
                                    (make-tile 4 25 9)
                                    (make-tile 1 26 9)
                                    (make-tile 4 27 9)
                                    (make-tile 1 28 9))))

; get-tile-image: tile -> image
; returns the image corresponding to the tile given
(define (get-tile-image t) (list-ref useable-tiles (tile-id t)))

(check-expect (get-tile-image (make-tile 0 0 0)) (crop-at-xy 0 0 tileset-mario1))
(check-expect (get-tile-image (make-tile 2 3 5)) (crop-at-xy 2 0 tileset-mario1))


; tile->renderpos: tile number -> posn
; takes a tile and the x position of the player and returns the posn where it should be rendered on screen
(define (tile->renderpos tile x)
  (make-posn (- (+ (* 16 (tile-x tile)) 8 (/ width 2)) x) (+ (* 16 (tile-y tile)) 8)))

(check-expect (tile->renderpos (make-tile 0 0 0) 0) (make-posn 136 8))
(check-expect (tile->renderpos (make-tile 0 2 5) 64) (make-posn 104 88))
(check-expect (tile->renderpos (make-tile 3 1 2) -122) (make-posn 274 40))


; render-tiles: list of tiles, image, number -> image
; renders all the tiles in the given list on top of the given background image with respect to the given player x location
(define (render-tiles li bg x)
  (let ([vis-tiles (filter (lambda (a) (and (>= (* (tile-x a) 16) (- x (/ width 2) 16)) (<= (* (tile-x a) 16) (+ x (/ width 2))))) li)])
    (place-images (map get-tile-image vis-tiles) (map (lambda (l) (tile->renderpos l x)) vis-tiles) bg)))


; first-predicate: [X -> boolean], list of X -> X
; takes a list of X and returns the first value where the given function returns true

(check-expect (first-predicate odd? (list 1 2 3)) 1)
(check-expect (first-predicate odd? (list 2 3 4)) 3)
(check-expect (first-predicate odd? (list 1)) 1)
(check-error (first-predicate odd? empty) "Empty list given or no items that match #<procedure:odd?> in given list")
(check-error (first-predicate odd? (list 2 4 6)) "Empty list given or no items that match #<procedure:odd?> in given list")

(check-expect (first-predicate even? (list 1 2 3)) 2)
(check-expect (first-predicate even? (list 2 3 4)) 2)
(check-expect (first-predicate even? (list 4)) 4)
(check-error (first-predicate even? empty) "Empty list given or no items that match #<procedure:even?> in given list")
(check-error (first-predicate even? (list 1 3 5)) "Empty list given or no items that match #<procedure:even?> in given list")

(define (first-predicate p li)
  (cond [(empty? li) (error (format "Empty list given or no items that match ~a in given list" p))]
        [(p (first li)) (first li)]
        [else (first-predicate p (rest li))]))


; tile-at-xy: non-empty list of tiles, integer, integer -> tile
; returns the first tile at the given x and y position in the list of tiles
(define (tile-at-xy li x y)
  (first-predicate (lambda (a) (and (= (tile-x a) x) (= (tile-y a) y))) li))

(check-expect (tile-at-xy (list (make-tile 0 0 0) (make-tile 0 1 3) (make-tile 3 2 4)) 0 0) (make-tile 0 0 0))
(check-expect (tile-at-xy (list (make-tile 0 0 0) (make-tile 0 1 3) (make-tile 3 2 4)) 1 3) (make-tile 0 1 3))
(check-expect (tile-at-xy (list (make-tile 3 2 4)) 2 4) (make-tile 3 2 4))


; player spriteset image
(define player-spriteset (bitmap/file "res/spriteset-mario1.png"))

; list of useable sprites
(define useable-sprites (list (crop 80 0 16 32 player-spriteset)
                              (crop 96 0 16 32 player-spriteset)
                              (crop 112 0 16 32 player-spriteset)
                              (crop 128 0 16 32 player-spriteset)))

; a sprite structure consists of:
; img: the id of the image to use for the sprite
; x: the x position of the sprite
; y: the y position of the sprite
; horiz-vel: the horizontal velocity of the sprite
; vert-vel: the vertical velocity of the sprite
(define-struct sprite (img x y horiz-vel vert-vel))

; get-sprite-image: sprite -> image
; takes a sprite struct and returns the image to use to render it
(define (get-sprite-image s) (list-ref useable-sprites (sprite-img s)))

(check-expect (get-sprite-image (make-sprite 0 0 0 0 0)) (crop 80 0 16 32 player-spriteset))
(check-expect (get-sprite-image (make-sprite 2 5 1 3 2)) (crop 112 0 16 32 player-spriteset))
(check-expect (get-sprite-image (make-sprite 0 2 5 1 7)) (crop 80 0 16 32 player-spriteset))


; the default player sprite to use in the program
(define default-player (make-sprite 0 128 128 0 0))


; worldstate structure consists of:
; mode: integer, current mode of the game (0: start screen, 1: play, 2: build)
; player: char struct, represents the player
; world-tiles: list (of tiles), represents the entire world as a list of tiles
; world-enemies: list (of sprites), represents enemies as a list of chars
; selected-tile: tile, represents the currently selected tile in the game
(define-struct ws (mode player tiles enemies selected-tile-id))

; the default worlstate to use when starting the program
(define default-ws (make-ws 0 default-player default-tiles empty 0))

; main: worldstate -> worldstate
; starts the program with big-bang
; takes the starting state (worldstate) of the program
; and returns the worldstate when it has finished
(define (main start)
  (big-bang start
            [to-draw render] 
            [on-tick tick]
            [on-mouse click]
            [on-key key]
            [name "horrible game"]
            ))


(define width 256) ; width of the scene

(define height-play 256) ; height of the play area of the game
(define height-top 32); height of the top area of the game

(define scale-factor 2) ; factor by which to scale the output


; remove-tile-xy: list of tiles, number, number -> list of tiles
; returns a list of tiles with any with the same x and y value as the given x and y value removed
(define (remove-tile-xy li x y)
  (cond [(empty? li) empty]
        [(and (= (tile-x (first li)) x) (= (tile-y (first li)) y)) (rest li)]
        [else (cons (first li) (remove-tile-xy (rest li) x y))]))

(check-expect (remove-tile-xy (list (make-tile 0 0 0) (make-tile 0 1 3) (make-tile 3 2 4)) 0 0) (list (make-tile 0 1 3) (make-tile 3 2 4)))
(check-expect (remove-tile-xy empty 0 0) empty)
(check-expect (remove-tile-xy (list (make-tile 0 0 0) (make-tile 0 1 3) (make-tile 3 2 4)) 8 8) (list (make-tile 0 0 0) (make-tile 0 1 3) (make-tile 3 2 4)))


; change-tile-id: list of tiles, number, number, integer, number -> list of tiles
; takes a list of tiles and returns one with the tile at the given screen x and y values
; with respect to the given player x value replaced with a tile at the same x and y value but
; with the given id
(define (change-tile-id li x y id px)
  (let ([tx (round (/ (+ (- (/ x scale-factor) 8 (/ width 2)) px) 16))]
        [ty (round (/ (- (/ y scale-factor) 8 height-top) 16))])
    (cons (make-tile id tx ty) (remove-tile-xy li tx ty))))

(check-expect (change-tile-id (list (make-tile 0 0 0) (make-tile 0 1 3) (make-tile 3 2 4)) 0 0 1 0) (list (make-tile 1 -8 -2) (make-tile 0 0 0) (make-tile 0 1 3) (make-tile 3 2 4)))
(check-expect (change-tile-id (list (make-tile 0 -8 -2) (make-tile 0 1 3) (make-tile 3 2 4)) 0 0 1 0) (list (make-tile 1 -8 -2) (make-tile 0 1 3) (make-tile 3 2 4)))


(define blank (rectangle width (+ height-play height-top) "solid" (make-color 93 148 251))) ; blank scene that everything is drawn on

; standard-text: string, number, string (color) -> image
; gives the text image of the provided string, at the provided size,
; in the provided color
; its job is to round the size and make sure it stays at size 255 or lower
; so it doesn't cause an error in the normal text function
(define (standard-text str size color)
  (text str (min (round size) 255) color))


; change-player: worldstate sprite -> worldstate
; simply replaces the player in the given worldstate with the given player
(define (change-player ws player)
  (make-ws (ws-mode ws) player (ws-tiles ws) (ws-enemies ws) (ws-selected-tile-id ws)))

(check-expect (change-player (make-ws 0 default-player default-tiles empty 0) (make-sprite 0 0 0 0 0)) (make-ws 0 (make-sprite 0 0 0 0 0) default-tiles empty 0))


; is-sprite-on-tile: sprite, tile -> boolean
; returns true if the sprite is on (or somewhat in) the given tile
(define (is-sprite-on-tile s t)
  (and (>= (sprite-x s) (* (tile-x t) 16)) (<= (sprite-x s) (* (+ (tile-x t) 1) 16))
       (>= (sprite-y s) (- (* (tile-y t) 16) 48)) (<= (sprite-y s) (- (* (tile-y t) 16) 16))))

(check-expect (is-sprite-on-tile (make-sprite 0 0 0 0 0) (make-tile 0 0 0)) #false)
(check-expect (is-sprite-on-tile (make-sprite 0 0 -48 0 0) (make-tile 0 0 0)) #true)


; is-sprite-on-tiles: sprite, list of tiles -> boolean
; returns true if the given sprite is on (or somewhat in) any tile in the given list
(define (is-sprite-on-tiles s li)
  (ormap (lambda (x) (is-sprite-on-tile s x)) li))

(check-expect (is-sprite-on-tiles (make-sprite 0 0 0 0 0) (list (make-tile 0 0 0) (make-tile 0 5 1))) #false)
(check-expect (is-sprite-on-tiles (make-sprite 0 0 -48 0 0) (list (make-tile 0 0 0) (make-tile 0 5 1))) #true)
(check-expect (is-sprite-on-tiles (make-sprite 0 0 0 0 0) empty) #false)


; tile-sprite-on: sprite, non-empty list of tiles -> tile
; takes a non-empty list of tiles which does contain a tile that the given sprite is on
; and returns the tile that the given sprite is on
(define (tile-sprite-on s li)
  (first-predicate (lambda (x) (is-sprite-on-tile s x)) li))

(check-expect (tile-sprite-on (make-sprite 0 0 -48 0 0) (list (make-tile 0 0 0) (make-tile 0 5 1))) (make-tile 0 0 0))


; calculate-sprite-physics: sprite, list of tiles -> sprite
; takes a sprite and the current list of tiles
; and calculates various physics on it like gravity and velocity in all directions
(define (calculate-sprite-physics s li)
  (let* ([isot (is-sprite-on-tiles s li)]
         [hv (cond [(= (sprite-horiz-vel s) 0) 0]
                   [(> (sprite-horiz-vel s) 0) (min (- (sprite-horiz-vel s) 1) 8)]
                   [(< (sprite-horiz-vel s) 0) (max (+ (sprite-horiz-vel s) 1) -8)])]
         [vv (if isot (min (sprite-vert-vel s) 0) (min (+ (sprite-vert-vel s) 1) 8))])
    (make-sprite (sprite-img s) (+ (sprite-x s) hv) (+ (if isot (- (* 16 (tile-y (tile-sprite-on s li))) 48) (sprite-y s)) vv) hv vv)))


; render: worldstate -> image
; takes the worldstate and puts together the final displayed image
(define (render ws)
  (scale scale-factor (cond [(= (ws-mode ws) 0) (place-images (list (standard-text "Instructions:" 24 "black")
                                                                    (standard-text "Press enter to start and change game mode" 10 "black")
                                                                    (standard-text "In play mode, press left/right/a/d to move" 10 "black")
                                                                    (standard-text "Also, press space space or click mouse to jump" 10 "black")
                                                                    (standard-text "In build mode, press a number to select tile" 10 "black")
                                                                    (standard-text "In build mode, click to palce the selected tile" 10 "black"))
                                                              (list (make-posn (/ width 2) 16)
                                                                    (make-posn (/ width 2) 48)
                                                                    (make-posn (/ width 2) 64)
                                                                    (make-posn (/ width 2) 80)
                                                                    (make-posn (/ width 2) 96)
                                                                    (make-posn (/ width 2) 112)) blank)]
                            [else (place-images (list (if (= (ws-mode ws) 1) (place-image (standard-text "Play Mode" 16 "black") (/ width 2) 8 (rectangle width height-top 0 "transparent"))
                                                          (place-images (list (standard-text "Edit Mode" 16 "black")
                                                                              (apply beside (build-list num-useable-tiles (lambda (x) (overlay (standard-text (number->string (+ x 1)) 16 (if (= x (ws-selected-tile-id ws)) "yellow" "white")) (list-ref useable-tiles x))))))
                                                                        (list (make-posn (/ width 2) 8)
                                                                              (make-posn (/ width 2) (- height-top 8)))
                                                                        (rectangle width height-top 0 "transparent")))
                                                      (place-image (get-sprite-image (ws-player ws))
                                                                   (/ width 2)
                                                                   (+ (sprite-y (ws-player ws)) height-top)
                                                                   (render-tiles (ws-tiles ws) (rectangle width height-play 0 "transparent") (sprite-x (ws-player ws)))))
                                                (list (make-posn (/ width 2) (/ height-top 2))
                                                      (make-posn (/ width 2) (+ (/ height-play 2) height-top)))
                                                blank)])))


; tick: worldstate -> worldstate
; takes the worldstate and if the game mode is 1, runs the physics on the player
; otherwise returns the given worldstate unchanged
(define (tick ws)
  (cond [(= (ws-mode ws) 1) (change-player ws (calculate-sprite-physics (ws-player ws) (ws-tiles ws)))]
        [else ws]))


; click: worldstate, number, number, string (mouse-event) -> worldstate
; if the game mode is 1, jumps when the mouse event is button-up
; if the game mode is 2, sets the clicked tile to the selected tile id if the mouse event is button-up
; otherwise returns the worldstate
(define (click ws x y type)
  (cond
    [(= (ws-mode ws) 1) (change-player ws (make-sprite (sprite-img (ws-player ws))
                                                       (sprite-x (ws-player ws)) 
                                                       (sprite-y (ws-player ws))
                                                       (sprite-horiz-vel (ws-player ws))
                                                       (+ (sprite-vert-vel (ws-player ws)) (cond [(and (mouse=? type "button-up") (is-sprite-on-tiles (ws-player ws) (ws-tiles ws)) (gui-play-sound jump-sound play-asynchronously)) -8]
                                                                                                 [else 0]))))]
    [(and (= (ws-mode ws) 2) (mouse=? type "button-up")) (make-ws (ws-mode ws)
                                                                  (ws-player ws)
                                                                  (change-tile-id (ws-tiles ws) x y (ws-selected-tile-id ws) (sprite-x (ws-player ws)))
                                                                  (ws-enemies ws)
                                                                  (ws-selected-tile-id ws))]
    [else ws]))


; key: worldstate, string -> worldstate
; if the key pressed is return, changes the game mode
; if the key pressed is numeric and is a valid useable tile, sets the current selected tile to it's value
; if the key pressed is up/down/left/right/w/a/s/d and the game mode is 2, moves the player
; if the key pressed is left/right/a/s/space and the game mode is 12, moves the player
; otherwise returns the worldstate
(define (key ws key-pressed)
  (cond [(and (key=? key-pressed "\r") (gui-play-sound pause-sound play-asynchronously)) (make-ws (if (= (ws-mode ws) 1) 2 1)
                                                                                                  (ws-player ws)
                                                                                                  (ws-tiles ws)
                                                                                                  (ws-enemies ws)
                                                                                                  (ws-selected-tile-id ws))]
        [(= (ws-mode ws) 0) ws]
        [(and (= (ws-mode ws) 2) (string-numeric? key-pressed)) (make-ws (ws-mode ws)
                                                                         (ws-player ws)
                                                                         (ws-tiles ws)
                                                                         (ws-enemies ws)
                                                                         (if (<= (string->number key-pressed) num-useable-tiles) (- (string->number key-pressed) 1) (ws-selected-tile-id ws)))]
        [(= (ws-mode ws) 2) (change-player ws (make-sprite (sprite-img (ws-player ws)) (+ (sprite-x (ws-player ws)) (cond [(or (key=? key-pressed "right") (key=? key-pressed "d")) 8]
                                                                                                                          [(or (key=? key-pressed "left") (key=? key-pressed "a")) -8]
                                                                                                                          [else 0]))
                                                           (+ (sprite-y (ws-player ws)) (cond [(or (key=? key-pressed "up") (key=? key-pressed "w")) -8]
                                                                                              [(or (key=? key-pressed "down") (key=? key-pressed "s")) 8]
                                                                                              [else 0]))
                                                           (sprite-horiz-vel (ws-player ws))
                                                           (sprite-vert-vel (ws-player ws))))]
        [(= (ws-mode ws) 1)  (change-player ws (make-sprite (sprite-img (ws-player ws))
                                                            (sprite-x (ws-player ws)) 
                                                            (sprite-y (ws-player ws))
                                                            (+ (sprite-horiz-vel (ws-player ws)) (cond [(or (key=? key-pressed "right") (key=? key-pressed "d")) 4]
                                                                                                       [(or (key=? key-pressed "left") (key=? key-pressed "a")) -4]
                                                                                                       [else 0]))
                                                            (+ (sprite-vert-vel (ws-player ws)) (cond [(and (key=? key-pressed " ") (is-sprite-on-tiles (ws-player ws) (ws-tiles ws)) (gui-play-sound jump-sound play-asynchronously)) -8]
                                                                                                      [else 0]))))]
        
        [else ws]))


; start the program
(main default-ws)
